import json, os, random, requests
from flask import (
    Blueprint,
    Flask,
    jsonify,
    url_for,
    redirect,
    render_template,
    request,
    flash,
)
from flask import current_app as app
from flask_cors import CORS
from .hwi_rpc import HWIBridge
from .helpers import deep_update, hwi_get_config, save_hwi_bridge_config


hwi_server = Blueprint("hwi_server", __name__)
CORS(hwi_server)
rand = random.randint(0, 1e32)  # to force style refresh
hwi = HWIBridge()


@hwi_server.route("/", methods=["GET"])
def index():
    return redirect(url_for("hwi_server.hwi_bridge_settings"))


@hwi_server.route("/api/", methods=["POST"])
def api():
    """JSON-RPC for HWI Bridge"""
    # if cross-origin
    if (
        "HTTP_HOST" in request.environ
        and "HTTP_ORIGIN" in request.environ
        and request.environ["HTTP_HOST"]
        != request.environ["HTTP_ORIGIN"].split("://")[1]
    ):
        whitelisted_domains = hwi_get_config(app.specter)["whitelisted_domains"].split()
        for i, url in enumerate(whitelisted_domains):
            # might be https as well
            whitelisted_domains[i] = url.replace("://localhost:", "://127.0.0.1:")
        if "*" not in whitelisted_domains:
            origin_url = request.environ["HTTP_ORIGIN"].replace(
                "://localhost:", "://127.0.0.1:"
            )
            if not origin_url.endswith("/"):
                # make sure the url end with a "/"
                origin_url += "/"
            if not (origin_url in whitelisted_domains):
                return (
                    jsonify(
                        {
                            "jsonrpc": "2.0",
                            "error": {
                                "code": -32001,
                                "message": "Unauthorized request origin.<br>You must first whitelist this website URL in HWIBridge settings to grant it access.",
                            },
                            "id": None,
                        }
                    ),
                    500,
                )
    try:
        data = json.loads(request.data)
    except:
        return (
            jsonify(
                {
                    "jsonrpc": "2.0",
                    "error": {"code": -32700, "message": "Parse error"},
                    "id": None,
                }
            ),
            500,
        )
    if ("forwarded_request" not in data or not data["forwarded_request"]) and (
        app.specter.hwi_bridge_url.startswith("http://")
        or app.specter.hwi_bridge_url.startswith("https://")
    ):
        if "HTTP_ORIGIN" not in request.environ:
            return (
                jsonify(
                    {
                        "jsonrpc": "2.0",
                        "error": {
                            "code": -32600,
                            "message": "Request must specify its origin or set `forwarded_request` to `true`.",
                        },
                        "id": None,
                    }
                ),
                500,
            )
        data["forwarded_request"] = True
        requests_session = app.specter.requests_session(
            force_tor=".onion/" in app.specter.hwi_bridge_url
        )
        requests_session.headers.update({"origin": request.environ["HTTP_ORIGIN"]})
        forwarded_request = requests_session.post(
            app.specter.hwi_bridge_url, data=json.dumps(data)
        )
        response = json.loads(forwarded_request.content)
        return jsonify(response)

    return jsonify(hwi.jsonrpc(data))


@hwi_server.route("/settings/", methods=["GET", "POST"])
def hwi_bridge_settings():
    config = hwi_get_config(app.specter)
    if request.method == "POST":
        action = request.form["action"]
        if action == "update":
            config["whitelisted_domains"] = request.form["whitelisted_domains"]
            save_hwi_bridge_config(app.specter, config)
            flash("Whitelist is updated!")
    return render_template(
        "hwi_bridge.jinja",
        specter=app.specter,
        whitelisted_domains=config["whitelisted_domains"],
        rand=rand,
    )



from flask import (
    Blueprint,
    request,
)
from flask_login import login_required, current_user
from flask import current_app as app
from ..util.price_providers import update_price

# Setup endpoint blueprint
price_endpoint = Blueprint("price_endpoint", __name__)

################ Price settings ####################
@price_endpoint.route("/update/", methods=["GET", "POST"])
@login_required
def update():
    try:
        price_type = request.form.get("price_type", "manual")
        if price_type == "manual":
            alt_rate = request.form.get("alt_rate", 0)
            alt_symbol = request.form.get("alt_symbol", "")
            app.specter.update_price_provider("", current_user)
            app.specter.price_checker.stop()
            if alt_rate and alt_symbol:
                app.specter.update_alt_rate(alt_rate, current_user)
                app.specter.update_alt_symbol(alt_symbol, current_user)
                return {"success": True}
        else:
            price_provider = request.form.get("price_provider", "")
            weight_unit = request.form.get("weight_unit", "oz")
            app.specter.update_price_provider(price_provider, current_user)
            app.specter.update_weight_unit(weight_unit, current_user)
            if not app.specter.price_checker.running:
                app.specter.price_checker.start()
            return {"success": update_price(app.specter, current_user)}
    except Exception as e:
        app.logger.warning("Failed to update price settings. Exception: {}".format(e))
    return {"success": False}


@price_endpoint.route("/toggle/", methods=["GET", "POST"])
@login_required
def toggle():
    try:
        app.specter.update_price_check_setting(
            not app.specter.price_check, current_user
        )
        return {"success": True}
    except Exception as e:
        app.logger.warning("Failed to update price settings. Exception: {}".format(e))
    return {"success": False}



import json
import logging
import os

from embit.liquid.networks import get_network
from flask_babel import lazy_gettext as _
from .helpers import is_testnet, is_liquid
from .liquid.rpc import LiquidRPC
from .persistence import write_node
from .rpc import (
    BitcoinRPC,
    RpcError,
    autodetect_rpc_confs,
    detect_rpc_confs,
    get_default_datadir,
)

logger = logging.getLogger(__name__)


class Node:
    """A Node represents the connection to a Bitcoin and/or Liquid (Full-) node.
    It can be created via Constructor or from_json, and mainly it can give you A
    RPC-object to use the API.
    On top of the RPC-connection it manages the stability of the rpc. It will only
    persist healthy connections.
    One or many Nodes are managed via the NodeManager
    """

    def __init__(
        self,
        name,
        alias,
        autodetect,
        datadir,
        user,
        password,
        port,
        host,
        protocol,
        external_node,
        fullpath,
        manager,
    ):
        """Constructor for your Node.

        :param name: arbitrary name
        :param alias: Bad habit, doesn't seem to have business functionality
        :param autodetect: Boolean, will use the datadir to derive config is yes
        :param datadir: A directory where a bitcoin.conf can be found, relevant for autodetect
        :param user: rpc-user
        :param password: rpc-password
        :param port: usually something like 8332 for mainnet, 18332 for testnet, 18443 for Regtest, 38332 for signet
        :param host: domainname or ip-address. Don't add the protocol here
        :param protocol: Usually https or http
        :param external_node: should be True for Node and False for InternalNode
        :param fullpath: it's assumed that you want to store it on disk AND decide about the fullpath upfront
        :param manager: A NodeManager instance which will get notified if the Node's name changes, the proxy_url will get copied from the manager as well
        """
        self.name = name
        self.alias = alias
        self.autodetect = autodetect
        self.datadir = datadir
        self.user = user
        self.password = password
        self.port = port
        self.host = host
        self.protocol = protocol
        self.external_node = external_node
        self.fullpath = fullpath
        self.manager = manager
        self.proxy_url = manager.proxy_url
        self.only_tor = manager.only_tor
        self.rpc = self.get_rpc()
        self._asset_labels = None

        self.check_info()

    @classmethod
    def from_json(cls, node_dict, manager, default_alias="", default_fullpath=""):
        """Create a Node from json"""
        name = node_dict.get("name", "")
        alias = node_dict.get("alias", default_alias)
        autodetect = node_dict.get("autodetect", True)
        datadir = node_dict.get("datadir", get_default_datadir())
        user = node_dict.get("user", "")
        password = node_dict.get("password", "")
        port = node_dict.get("port", None)
        host = node_dict.get("host", "localhost")
        protocol = node_dict.get("protocol", "http")
        external_node = node_dict.get("external_node", True)
        fullpath = node_dict.get("fullpath", default_fullpath)

        return cls(
            name,
            alias,
            autodetect,
            datadir,
            user,
            password,
            port,
            host,
            protocol,
            external_node,
            fullpath,
            manager,
        )

    @property
    def json(self):
        """Get a json-representation of this Node"""
        return {
            "name": self.name,
            "alias": self.alias,
            "autodetect": self.autodetect,
            "datadir": self.datadir,
            "user": self.user,
            "password": self.password,
            "port": self.port,
            "host": self.host,
            "protocol": self.protocol,
            "external_node": self.external_node,
            "fullpath": self.fullpath,
        }

    def get_rpc(self):
        """
        Checks if config have changed, compares with old rpc
        and returns new one if necessary
        """
        if hasattr(self, "rpc"):
            rpc = self.rpc
        else:
            rpc = None
        if self.autodetect:
            if self.port:
                rpc_conf_arr = autodetect_rpc_confs(
                    datadir=os.path.expanduser(self.datadir), port=self.port
                )
            else:
                rpc_conf_arr = autodetect_rpc_confs(
                    datadir=os.path.expanduser(self.datadir)
                )
            if len(rpc_conf_arr) > 0:
                rpc = BitcoinRPC(
                    **rpc_conf_arr[0], proxy_url=self.proxy_url, only_tor=self.only_tor
                )
            if rpc == None:
                logger.warning(f"No rpc was found for {self}")
            return rpc
        else:
            # if autodetect is disabled and port is not defined
            # we use default port 8332
            if not self.port:
                self.port = 8332
            rpc = BitcoinRPC(
                user=self.user,
                password=self.password,
                host=self.host,
                port=self.port,
                protocol=self.protocol,
                proxy_url=self.proxy_url,
                only_tor=self.only_tor,
            )

        if rpc == None:
            logger.error(f"connection results to None in get_rpc")
            return None
        # check if it's liquid
        try:
            res = rpc.getblockchaininfo()
            if is_liquid(res.get("chain")):
                # convert to LiquidRPC class
                rpc = LiquidRPC.from_bitcoin_rpc(rpc)
        except RpcError as rpce:
            if rpce.status_code == 401:
                return rpc  # The user is failing to configure correctly
            logger.exception(rpce)
            return None
        except Exception as e:
            logger.exception(e)
            return None
        if rpc.test_connection():
            return rpc
        else:
            logger.debug(
                f"connection {rpc} fails test_connection() returning None in get_rpc"
            )
            return None

    def update_rpc(
        self,
        autodetect=None,
        datadir=None,
        user=None,
        password=None,
        port=None,
        host=None,
        protocol=None,
    ):
        """Changes the attributes of that node but only persists it, if the rpc.test_connection succeeds"""
        update_rpc = self.rpc is None or not self.rpc.test_connection()
        if autodetect is not None and self.autodetect != autodetect:
            logger.debug(f"{self} updating autodetect to {autodetect}")
            self.autodetect = autodetect
            update_rpc = True
        if datadir is not None and self.datadir != datadir:
            logger.debug(f"{self} updating datadir to {datadir}")
            self.datadir = datadir
            update_rpc = True
        if user is not None and self.user != user:
            logger.debug(f"{self} updating user to {user}")
            self.user = user
            update_rpc = True
        if password is not None and self.password != password:
            logger.debug(f"{self} updating password to XXXXXXXX")
            self.password = password
            update_rpc = True
        if port is not None and self.port != port:
            logger.debug(f"{self} updating port to {port}")
            self.port = port
            update_rpc = True
        if host is not None and self.host != host:
            logger.debug(f"{self} updating host to {host}")
            self.host = host
            update_rpc = True
        if protocol is not None and self.protocol != protocol:
            logger.debug(f"{self} updating protocol to {protocol}")
            self.protocol = protocol
            update_rpc = True
        if update_rpc:
            self.rpc = self.get_rpc()
            if self.rpc and self.rpc.test_connection():
                logger.debug(f"persisting {self} in update_rpc")
                write_node(self, self.fullpath)
        self.check_info()
        return False if not self.rpc else self.rpc.test_connection()

    def rename(self, new_name):
        logger.info("Renaming {}".format(self.alias))
        self.name = new_name
        logger.info(f"persisting {self} in rename")
        write_node(self, self.fullpath)
        self.manager.update()

    def check_info(self):
        self._is_configured = self.rpc is not None
        self._is_running = False
        if self.rpc is not None and self.rpc.test_connection():
            try:
                res = [
                    r["result"]
                    for r in self.rpc.multi(
                        [
                            ("getblockchaininfo", None),
                            ("getnetworkinfo", None),
                            ("getmempoolinfo", None),
                            ("uptime", None),
                            ("getblockhash", 0),
                            ("scantxoutset", "status", []),
                        ]
                    )
                ]
                self._info = res[0]
                self._network_info = res[1]
                self._info["mempool_info"] = res[2]
                self._info["uptime"] = res[3]
                try:
                    self.rpc.getblockfilter(res[4])
                    self._info["blockfilterindex"] = True
                except:
                    self._info["blockfilterindex"] = False
                self._info["utxorescan"] = (
                    res[5]["progress"]
                    if res[5] is not None and "progress" in res[5]
                    else None
                )
                if self._info["utxorescan"] is None:
                    self.utxorescanwallet = None
                self._network_parameters = get_network(self.chain)
                self._is_running = True
            except Exception as e:
                self._info = {"chain": None}
                self._network_info = {"subversion": "", "version": 999999}
                self._network_parameters = get_network("main")
                logger.error(f"connection {self.rpc} could not suceed check_info")
                logger.exception("Exception %s while check_info()" % e)
        else:
            if self.rpc is None:
                logger.warning(f"connection of {self} is None in check_info")
            elif not self.rpc.test_connection():
                logger.debug(
                    f"connection {self.rpc} failed test_connection in check_info:"
                )
                try:
                    self.rpc.multi(
                        [
                            ("getblockchaininfo", None),
                            ("getnetworkinfo", None),
                            ("getmempoolinfo", None),
                            ("uptime", None),
                            ("getblockhash", 0),
                            ("scantxoutset", "status", []),
                        ]
                    )
                except Exception as e:
                    logger.error(e)
            self._info = {"chain": None}
            self._network_info = {"subversion": "", "version": 999999}

        if not self._is_running:
            self._info["chain"] = None

    def test_rpc(self):
        """tests the rpc-connection and returns a dict which helps
        to derive what might be wrong with the config
        ToDo: list an example here.
        """
        rpc = self.get_rpc()
        if rpc is None:
            return {
                "out": "",
                "err": _("Connection to node failed"),
                "code": -1,
                "tests": {},
            }
        r = {}
        r["tests"] = {"connectable": False}
        r["err"] = ""
        r["code"] = 0
        try:
            r["tests"]["recent_version"] = (
                int(rpc.getnetworkinfo()["version"]) >= 200000
            )
            if not r["tests"]["recent_version"]:
                r["err"] = _("Core Node might be too old")

            r["tests"]["connectable"] = True
            r["tests"]["credentials"] = True
            try:
                rpc.listwallets()
                r["tests"]["wallets"] = True
            except RpcError as rpce:
                logger.info(f"Couldn't list wallets while test_rpc {rpce}")
                r["tests"]["wallets"] = False
                r["err"] = "Wallets disabled"

            r["out"] = json.dumps(rpc.getblockchaininfo(), indent=4)
        except ConnectionError as e:
            logger.info("Caught an ConnectionError while test_rpc: %s", e)

            r["tests"]["connectable"] = False
            r["err"] = _("Failed to connect!")
            r["code"] = -1
        except RpcError as rpce:
            logger.info(
                f"Caught an RpcError while test_rpc status_code: {rpce.status_code} error_code:{rpce.error_code}"
            )
            r["tests"]["connectable"] = True
            r["code"] = rpc.r.status_code
            if rpce.status_code == 401:
                r["tests"]["credentials"] = False
                r["err"] = _("RPC authentication failed!")
            else:
                r["err"] = str(rpce.status_code)
        except Exception as e:
            logger.error(
                "Caught an exception of type {} while test_rpc: {}".format(
                    type(e), str(e)
                )
            )
            r["out"] = ""
            if rpc.r is not None and "error" in rpc.r:
                r["err"] = rpc.r["error"]
                r["code"] = rpc.r.status_code
            else:
                r["err"] = _("Failed to connect")
                r["code"] = -1
        return r

    def abortrescanutxo(self):
        """use this to abort a rescan as it stores some state while rescanning"""
        self.rpc.scantxoutset("abort", [])
        # Bitcoin Core doesn't catch up right away
        # so app.specter.check() doesn't work
        self._info["utxorescan"] = None
        self.utxorescanwallet = None

    def check_blockheight(self):
        return self.info["blocks"] != self.rpc.getblockcount()

    def is_liquid(self):
        return is_liquid(self.chain)

    @property
    def is_running(self):
        return self._is_running

    @property
    def is_configured(self):
        return self._is_configured

    @property
    def info(self):
        return self._info

    @property
    def network_info(self):
        return self._network_info

    @property
    def network_parameters(self):
        try:
            return self._network_parameters
        except Exception:
            return get_network("main")

    @property
    def bitcoin_core_version(self):
        # return self.network_info["subversion"].replace("/", "").replace("Satoshi:", "")
        # This hopefully works for elements as well:
        return self.network_info["subversion"].strip("/").split(":")[-1]

    @property
    def bitcoin_core_version_raw(self):
        return self.network_info["version"]

    @property
    def taproot_support(self):
        try:
            # currently only master branch supports tr() descriptors
            # TODO: replace to 220000
            return (self.bitcoin_core_version_raw >= 219900) and (
                self.info.get("softforks", {}).get("taproot", {}).get("active", False)
            )
        except Exception as e:
            return False

    @property
    def chain(self):
        return self.info["chain"]

    @property
    def is_testnet(self):
        return is_testnet(self.chain)

    @property
    def asset_labels(self):
        if not self.is_liquid:
            return {}
        if self._asset_labels is None:
            asset_labels = self.rpc.dumpassetlabels()
            assets = {}
            LBTC = "LBTC" if self.chain == "liquidv1" else "tLBTC"
            for k in asset_labels:
                assets[asset_labels[k]] = k if k != "bitcoin" else LBTC
            self._asset_labels = assets
        return self._asset_labels

    @property
    def is_liquid(self):
        return is_liquid(self.chain)

    @property
    def rpc(self):
        if not hasattr(self, "_rpc"):
            return None
        else:
            return self._rpc

    @rpc.setter
    def rpc(self, value):
        if hasattr(self, "_rpc") and self._rpc != value:
            logger.debug(f"Updating {self}.rpc {self._rpc} with {value} (setter)")
        if hasattr(self, "_rpc") and value == None:
            logger.debug(f"Updating {self}.rpc {self._rpc} with None (setter)")
        self._rpc = value

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name} fullpath={self.fullpath}>"



import requests
import logging

logger = logging.getLogger(__name__)

OZ_TO_G = 28.3495231


def update_price(specter, current_user):
    success, price, symbol = get_price_at(specter, current_user, timestamp="now")
    if success:
        specter.update_alt_rate(price, current_user)
        specter.update_alt_symbol(symbol, current_user)
    return success


"""
    Tries to get the current BTC price based on the user provider preferences.
    Returns: (success, price, symbol)
"""


def get_price_at(specter, current_user, timestamp="now"):
    try:
        if specter.price_check:
            requests_session = specter.requests_session(
                force_tor=("spotbit" in specter.price_provider)
            )
            currency = "usd"
            currency_symbol = "$"
            weight_unit_convertible = False
            if specter.price_provider.endswith("_eur"):
                currency = "eur"
                currency_symbol = "€"
            elif specter.price_provider.endswith("_gbp"):
                currency = "gbp"
                currency_symbol = "£"
            elif specter.price_provider.endswith("_chf"):
                currency = "chf"
                currency_symbol = " Fr."
            elif specter.price_provider.endswith("_aud"):
                currency = "aud"
                currency_symbol = "$"
            elif specter.price_provider.endswith("_cad"):
                currency = "cad"
                currency_symbol = "$"
            elif specter.price_provider.endswith("_nzd"):
                currency = "nzd"
                currency_symbol = "$"
            elif specter.price_provider.endswith("_hkd"):
                currency = "hkd"
                currency_symbol = "$"
            elif specter.price_provider.endswith("_jpy"):
                currency = "jpy"
                currency_symbol = "¥"
            elif specter.price_provider.endswith("_rub"):
                currency = "rub"
                currency_symbol = "₽"
            elif specter.price_provider.endswith("_ils"):
                currency = "ils"
                currency_symbol = "₪"
            elif specter.price_provider.endswith("_jod"):
                currency = "jod"
                currency_symbol = "د.ا"
            elif specter.price_provider.endswith("_twd"):
                currency = "twd"
                currency_symbol = "$"
            elif specter.price_provider.endswith("_brl"):
                currency = "brl"
                currency_symbol = " BRL"
            elif specter.price_provider.endswith("_xau"):
                currency = "xau"
                currency_symbol = " oz. "
                weight_unit_convertible = True
            elif specter.price_provider.endswith("_xag"):
                currency = "xag"
                currency_symbol = " oz. "
                weight_unit_convertible = True
            elif specter.price_provider.endswith("_xpt"):
                currency = "xpt"
                currency_symbol = " oz. "
                weight_unit_convertible = True
            elif specter.price_provider.endswith("_xpd"):
                currency = "xpd"
                currency_symbol = " oz. "
                weight_unit_convertible = True

            if specter.price_provider.startswith("bitstamp"):
                if timestamp == "now":
                    price = requests_session.get(
                        "https://www.bitstamp.net/api/v2/ticker/btc{}".format(currency)
                    ).json()["last"]
                else:
                    price = requests_session.get(
                        "https://www.bitstamp.net/api/v2/ohlc/btc{}/?limit=1&step=86400&start={}".format(
                            currency, timestamp
                        )
                    ).json()["data"]["ohlc"][0]["close"]
            elif specter.price_provider.startswith("coindesk"):
                if timestamp == "now":
                    price = requests_session.get(
                        f"https://api.coindesk.com/v1/bpi/currentprice/{currency.upper()}.json"
                    ).json()["bpi"][currency.upper()]["rate_float"]
                else:
                    return False, 0, ""
            elif specter.price_provider.startswith("spotbit"):
                exchange = specter.price_provider.split("spotbit_")[1].split("_")[0]
                if timestamp == "now":
                    price = requests_session.get(
                        "http://h6zwwkcivy2hjys6xpinlnz2f74dsmvltzsd4xb42vinhlcaoe7fdeqd.onion/now/{}/{}".format(
                            currency, exchange
                        )
                    ).json()["close"]
                else:
                    price = requests_session.get(
                        "http://h6zwwkcivy2hjys6xpinlnz2f74dsmvltzsd4xb42vinhlcaoe7fdeqd.onion/hist/{}/{}/{}/{}".format(
                            currency,
                            exchange,
                            timestamp * 1000,
                            (timestamp + 121) * 1000,
                        )
                    ).json()["data"][0][7]
            if weight_unit_convertible:
                if specter.weight_unit == "gram":
                    price = price * OZ_TO_G
                    currency_symbol = " g."
                elif specter.weight_unit == "kg":
                    price = price * OZ_TO_G / 1000
                    currency_symbol = " kg"

            return (True, price, currency_symbol)
    except Exception as e:
        logger.warning(
            "Failed to get price data from: {}. Exception: {}".format(
                specter.price_provider, e
            )
        )
    return (False, 0, "")



from .device_types import DeviceTypes
from .coldcard import ColdCard
from .trezor import Trezor
from .ledger import Ledger

from .bitbox02 import BitBox02
from .keepkey import Keepkey
from .specter import Specter
from .keystone import Keystone
from .cobo import Cobo
from .passport import Passport
from .jade import Jade
from .generic import GenericDevice
from .electrum import Electrum
from .bitcoin_core import BitcoinCore
from .elements_core import ElementsCore
from .seedsigner import SeedSignerDevice

# all device types
__all__ = [
    Trezor,
    Ledger,
    BitBox02,
    Specter,
    ColdCard,
    Keepkey,
    Keystone,
    Cobo,
    Passport,
    Jade,
    SeedSignerDevice,
    Electrum,
    BitcoinCore,
    ElementsCore,
    GenericDevice,
]



from .node_controller import NodePlainController
import atexit
import logging
import signal

logger = logging.getLogger(__name__)


class BitcoindPlainController(NodePlainController):
    """A Controller specifically for the Bitcoind-process"""

    def __init__(
        self,
        bitcoind_path="bitcoind",
        rpcport=18443,
        network="regtest",
        rpcuser="bitcoin",
        rpcpassword="secret",
    ):
        # Just call super and add the node_impl
        super().__init__(
            "bitcoin",
            node_path=bitcoind_path,
            rpcport=rpcport,
            network=network,
            rpcuser=rpcuser,
            rpcpassword=rpcpassword,
        )

    def start_bitcoind(
        self,
        cleanup_at_exit=False,
        cleanup_hard=False,
        datadir=None,
        log_stdout=False,
        extra_args=[],
        timeout=60,
    ):
        """starts bitcoind with a specific rpcport=18543 by default.
        That's not the standard in order to make pytest running while
        developing locally against a different regtest-instance
        if bitcoind_path == docker, it'll run bitcoind via docker.
        Specify a longer timeout for slower devices (e.g. Raspberry Pi)
        """
        # convenience method
        return self.start_node(
            cleanup_at_exit,
            cleanup_hard,
            datadir,
            log_stdout,
            extra_args,
            timeout,
        )

    def attach_to_proc_id(self, bitcoind_process):
        """This assumes that the calling context (prob. a internal_node instance) found a process which is suited.
        So instead of starting, we're somehow faking a start.
        This behaviour should be explicitely triggered as the management of that should be duty of InternalNode
        """
        # avoid circular reference
        from ..internal_node import BitcoindProcess

        self.datadir = bitcoind_process.get_cmd_arg_value("datadir")
        self.node_proc = bitcoind_process
        self.cleanup_hard = (
            False  # assuming an internal node which should not get killed -9
        )

        def cleanup_node_callback(signal_number=None, stack=None):
            self.cleanup_node(False, self.datadir)

        atexit.register(cleanup_node_callback)
        # This is for CTRL-C --> SIGINT
        signal.signal(signal.SIGINT, cleanup_node_callback)
        # This is for kill $pid --> SIGTERM
        signal.signal(signal.SIGTERM, cleanup_node_callback)

        self.status = "Running"

    def stop_bitcoind(self):
        return self.stop_node()

    def version(self):
        """Returns the version of bitcoind, e.g. "v0.19.1" """
        version = self.get_rpc().getnetworkinfo()["subversion"]
        version = version.replace("/", "").replace("Satoshi:", "v")
        return version



import logging
import os
import platform
import signal
import socket
import subprocess
from io import StringIO

import psutil

from .specter_error import SpecterError

logger = logging.getLogger(__name__)


class TorDaemonController:
    """A class controlling the tor-daemon process directly on the machine"""

    def __init__(
        self,
        tor_daemon_path,
        tor_config_path,
    ):
        self.tor_daemon_path = tor_daemon_path
        self.tor_config_path = tor_config_path
        self.tor_daemon_proc = None
        self.is_running()  # throws an exception if port 9050 is open (which looks like another tor-process is running)

    def start_tor_daemon(self, cleanup_at_exit=True):
        if self.is_running():
            if self.tor_daemon_proc is None:
                raise Exception(
                    "Tor Daemon running but not via this Controller instance."
                )
            return

        self.tor_daemon_proc = subprocess.Popen(
            f'{"exec " if platform.system() != "Windows" else ""}"{self.tor_daemon_path}" --defaults-torrc {self.tor_config_path}',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        logger.debug(
            "Running tor-daemon process with pid {}".format(self.tor_daemon_proc.pid)
        )

    def get_logs(self):
        if self.is_running():
            return "Cannot return logs as tor is still running"
        logs = ""
        newline = self.tor_daemon_proc.stdout.readline().decode("ascii")
        while newline != "":
            logs = logs + newline
            newline = self.tor_daemon_proc.stdout.readline().decode("ascii")
        return logs

    def get_hashed_password(self, password):
        hashed_pw = subprocess.check_output(
            f'{"exec " if platform.system() != "Windows" else ""}"{self.tor_daemon_path}" --hash-password {password}',
            shell=True,
        ).decode("utf8")

        if not hashed_pw.startswith("16:"):
            return "16:" + hashed_pw.split("\n16:")[1]
        return hashed_pw

    def is_running(self):
        # This fails the cypress-tests, unfortunately
        # if not self.tor_daemon_proc and self.is_port_open():
        #    raise SpecterError(
        #        "Port 9050 is open but tor_daemon_proc is not existing. Probably another Tor-Daemon is running?!"
        #    )
        return self.tor_daemon_proc and self.tor_daemon_proc.poll() is None

    def is_port_open(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        location = ("127.0.0.1", 9050)
        try:
            s.connect(location)
            s.close()
            return True
        except Exception as e:
            return False

    def stop_tor_daemon(self):
        timeout = 50  # in secs
        if self.tor_daemon_proc:
            if platform.system() == "Windows":
                subprocess.run("Taskkill /IM tor.exe /F")
            self.tor_daemon_proc.terminate()
            procs = psutil.Process().children()
            for p in procs:
                p.terminate()
            _, alive = psutil.wait_procs(procs, timeout=timeout)
            for p in alive:
                logger.info("tor daemon did not terminated in time, killing!")
                p.kill()



from .hwi_device import HWIDevice


class SDCardDevice(HWIDevice):

    sd_card_support = True
    exportable_to_wallet = True

    def create_psbts(self, base64_psbt, wallet):
        psbts = super().create_psbts(base64_psbt, wallet)
        psbts["sdcard"] = wallet.fill_psbt(base64_psbt, non_witness=True, xpubs=True)
        return psbts



import os
import shutil
import hashlib
import binascii
import json
from flask_login import UserMixin
from .specter_error import SpecterError
from .persistence import read_json_file, write_json_file, delete_folder
from .managers.wallet_manager import WalletManager
from .managers.device_manager import DeviceManager
from .helpers import deep_update


def hash_password(password):
    """Hash a password for storing."""
    salt = binascii.b2a_base64(hashlib.sha256(os.urandom(60)).digest()).strip()
    pwdhash = (
        binascii.b2a_base64(
            hashlib.pbkdf2_hmac("sha256", password.encode("utf-8"), salt, 10000)
        )
        .strip()
        .decode()
    )
    return {"salt": salt.decode(), "pwdhash": pwdhash}


def verify_password(stored_password, provided_password):
    """Verify a stored password against one provided by user"""
    pwdhash = hashlib.pbkdf2_hmac(
        "sha256",
        provided_password.encode("utf-8"),
        stored_password["salt"].encode(),
        10000,
    )
    return pwdhash == binascii.a2b_base64(stored_password["pwdhash"])


class User(UserMixin):
    def __init__(self, id, username, password, config, specter, is_admin=False):
        self.id = id
        self.username = username
        self.password = password
        self.config = config
        self.is_admin = is_admin
        self.uid = specter.config["uid"]
        self.specter = specter
        self.wallet_manager = None
        self.device_manager = None
        self.manager = None

    @property
    def folder_id(self):
        if self.is_admin:
            return ""
        return f"_{self.id}"

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        """pass a json or a plain-password here"""
        try:
            if value.get("salt") and value.get("pwdhash"):
                self._password = value
        except:
            salted_hashed_password = hash_password(value)
            self._password = salted_hashed_password

    @classmethod
    def from_json(cls, user_dict, specter):
        # TODO: Unify admin in backwards compatible way
        try:
            if not user_dict["is_admin"]:
                return cls(
                    user_dict["id"],
                    user_dict["username"],
                    user_dict["password"],
                    user_dict["config"],
                    specter,
                )
            else:
                return cls(
                    user_dict["id"],
                    user_dict["username"],
                    user_dict["password"],
                    {},
                    specter,
                    is_admin=True,
                )
        except Exception as e:
            raise SpecterError(f"Unable to parse user JSON.:{e}")

    @property
    def json(self):
        user_dict = {
            "id": self.id,
            "username": self.username,
            "password": self.password,
            "is_admin": self.is_admin,
        }
        if not self.is_admin:
            user_dict["config"] = self.config
        return user_dict

    def check(self):
        self.check_device_manager()
        self.check_wallet_manager()

    def check_wallet_manager(self):
        """Updates wallet manager for this user"""
        # if chain, user or data folder changed
        wallet_manager = self.wallet_manager
        wallets_rpcpath = "specter%s" % self.uid
        wallets_folder = os.path.join(
            self.specter.data_folder, f"wallets{self.folder_id}"
        )
        if (
            wallet_manager is None
            or wallet_manager.data_folder != wallets_folder
            or wallet_manager.rpc_path != wallets_rpcpath
            or wallet_manager.chain != self.specter.chain
        ):

            wallet_manager = WalletManager(
                self.specter.bitcoin_core_version_raw,
                wallets_folder,
                self.specter.rpc,
                self.specter.chain,
                self.device_manager,
                path=wallets_rpcpath,
            )
            self.wallet_manager = wallet_manager
        else:
            wallet_manager.update(
                wallets_folder, self.specter.rpc, chain=self.specter.chain
            )

    def check_device_manager(self, user=None):
        """Updates device manager for this user"""
        devices_folder = os.path.join(
            self.specter.data_folder, f"devices{self.folder_id}"
        )
        if self.device_manager is None:
            self.device_manager = DeviceManager(devices_folder)
        else:
            self.device_manager.update(data_folder=devices_folder)

    def save_info(self, delete=False):
        if self.manager is None:
            self.manager = self.specter.user_manager
        users = self.manager.users
        existing = self in users

        # update specter users
        if not existing and not delete:
            self.specter.add_user(self)
        if existing and delete:
            self.specter.delete_user(self)
        self.manager.save()

    def update_asset_label(self, asset, label, chain):
        if "asset_labels" not in self.config:
            self.config["asset_labels"] = {}
        if not label:
            if self.config["asset_labels"].get(chain, {}).get(asset):
                del self.config["asset_labels"][chain][asset]
        else:
            deep_update(self.config["asset_labels"], {chain: {asset: label}})
        self.save_info()

    def set_explorer(self, explorer_id, explorer):
        if "explorers" not in self.config:
            self.config["explorers"] = (
                {"main": "", "test": "", "regtest": "", "signet": ""},
            )
        self.config["explorers"][self.specter.chain] = explorer
        if "explorer_id" not in self.config:
            self.config["explorer_id"] = {
                "main": "CUSTOM",
                "test": "CUSTOM",
                "regtest": "CUSTOM",
                "signet": "CUSTOM",
            }
        self.config["explorer_id"][self.specter.chain] = explorer_id
        self.save_info()

    def set_fee_estimator(self, fee_estimator, custom_url):
        self.config["fee_estimator"] = fee_estimator
        if fee_estimator == "custom":
            self.config["fee_estimator_custom_url"] = custom_url
        self.save_info()

    def set_hwi_bridge_url(self, url):
        self.config["hwi_bridge_url"] = url
        self.save_info()

    def set_unit(self, unit):
        self.config["unit"] = unit
        self.save_info()

    def set_price_check(self, price_check_bool):
        self.config["price_check"] = price_check_bool
        self.save_info()

    def set_hide_sensitive_info(self, hide_sensitive_info_bool):
        self.config["hide_sensitive_info"] = hide_sensitive_info_bool
        self.save_info()

    def set_price_provider(self, price_provider):
        self.config["price_provider"] = price_provider
        self.save_info()

    def set_weight_unit(self, weight_unit):
        self.config["weight_unit"] = weight_unit
        self.save_info()

    def set_alt_rate(self, alt_rate):
        self.config["alt_rate"] = alt_rate
        self.save_info()

    def set_alt_symbol(self, alt_symbol):
        self.config["alt_symbol"] = alt_symbol
        self.save_info()

    def delete(self):
        # we delete wallet manager and device manager in save_info
        self.save_info(delete=True)

    def __eq__(self, other):
        if other == None:
            return False
        if isinstance(other, str):
            return self.id == other
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        # to make lookups in dicts by user id
        return hash(self.id)

    def __str__(self):
        return self.id

    def __repr__(self):
        return f"User({self.id})"



import os, platform
from stem.control import Controller


def get_tor_daemon_suffix():
    if platform.system() == "Darwin":
        return ".real"
    elif platform.system() == "Windows":
        return ".exe"
    return ""


def start_hidden_service(app):
    app.specter.tor_controller.authenticate(
        password=app.specter.config.get("torrc_password", "")
    )
    key_path = os.path.abspath(
        os.path.join(app.specter.data_folder, ".tor_service_key")
    )
    app.tor_service_id = None

    if not os.path.exists(key_path):
        service = app.specter.tor_controller.create_ephemeral_hidden_service(
            {app.tor_port: app.port}, await_publication=True
        )
        app.tor_service_id = service.service_id
        print(
            "* Started a new hidden service with the address of %s.onion"
            % app.tor_service_id
        )

        with open(key_path, "w") as key_file:
            key_file.write("%s:%s" % (service.private_key_type, service.private_key))
    else:
        with open(key_path) as key_file:
            key_type, key_content = key_file.read().split(":", 1)

        service = app.specter.tor_controller.create_ephemeral_hidden_service(
            {app.tor_port: app.port},
            key_type=key_type,
            key_content=key_content,
            await_publication=True,
        )
        app.tor_service_id = service.service_id
        print("* Resumed %s.onion" % app.tor_service_id)

    # save address to file
    if app.save_tor_address_to is not None:
        with open(app.save_tor_address_to, "w") as f:
            f.write("%s.onion" % app.tor_service_id)
    app.tor_service_id = app.tor_service_id
    app.tor_enabled = True
    if app.specter.config["auth"].get("method", "none") == "none":
        print(" * ############################# Warning! #############################")
        print(
            " * Your are running Specter over Tor with no authentication settings configured."
        )
        print(
            " * This means your Specter instance is accessible to anyone with the .onion URL."
        )
        print(
            " * This .onion URL is publicly exposed and indexed on the Tor network - it is not secret!"
        )
        print(
            " * It is stronly adviced that you configure proper authentication while running Specter behind a Tor hidden service."
        )
        print(
            " * Please go to Settings -> Authentication and set up an authentication method."
        )
        print(" * ####################################################################")


def stop_hidden_services(app):
    try:
        app.specter.tor_controller.authenticate(
            password=app.specter.config.get("torrc_password", "")
        )
        hidden_services = app.specter.tor_controller.list_ephemeral_hidden_services()
        print(" * Shutting down our hidden service")
        for tor_service_id in hidden_services:
            app.specter.tor_controller.remove_ephemeral_hidden_service(tor_service_id)
        # Sanity
        if len(app.specter.tor_controller.list_ephemeral_hidden_services()) != 0:
            print(" * Failed to shut down our hidden services...")
        else:
            print(" * Hidden services were shut down successfully")
            app.tor_service_id = None
    except Exception:
        pass  # we tried...



from . import DeviceTypes
from .hwi_device import HWIDevice


class BitBox02(HWIDevice):
    device_type = DeviceTypes.BITBOX02
    name = "BitBox02"
    icon = "bitbox02_icon.svg"
    supports_hwi_multisig_display_address = True



import logging
import json

import cryptoadvance.specter as specter
from ...wallet import Wallet
from .base import (
    SecureResource,
    rest_resource,
)
from ..security import require_admin, verify_password
from flask_restful import abort
from flask import current_app as app
from datetime import datetime
from ...specter_error import SpecterError

from ...util.fee_estimation import get_fees

from .. import auth
from .resource_healthz import ResourceLiveness, ResourceReadyness
from .resource_psbt import ResourcePsbt
from .resource_specter import ResourceSpecter
from .resource_txlist import ResourceTXlist

logger = logging.getLogger(__name__)


@rest_resource
class ResourceWallet(SecureResource):

    endpoints = ["/v1alpha/wallets/<wallet_alias>/"]

    def get(self, wallet_alias):
        user = auth.current_user()
        try:
            wallet: Wallet = app.specter.user_manager.get_user(
                user
            ).wallet_manager.get_by_alias(wallet_alias)
        except SpecterError as se:
            logger.error(se)
            if str(se).startswith(f"Wallet {wallet_alias} does not exist!"):
                return abort(403, message=f"Wallet {wallet_alias} does not exist")
            logger.error(se)
            return abort(500)

        wallet.get_balance()
        wallet.check_utxo()
        wallet.check_unused()

        return_dict = {}
        address_index = wallet.address_index
        validate_merkle_proofs = app.specter.config.get("validate_merkle_proofs")
        tx_list = []
        idx = 0
        tx_len = 1
        transactions = app.specter.wallet_manager.full_txlist(
            fetch_transactions=False, validate_merkle_proofs=validate_merkle_proofs
        )
        tx_list.append(transactions)

        # Flatten the list
        flat_list = []
        for element in tx_list:
            for dic_item in element:
                flat_list.append(dic_item)

        # Check if scanning
        scan = wallet.rescan_progress
        return_dict[wallet_alias] = wallet.__dict__
        return_dict["txlist"] = flat_list
        return_dict["scan"] = scan
        return_dict["address_index"] = address_index
        return_dict["utxo"] = wallet.utxo

        # Serialize only objects that are json compatible
        # This will exclude classes and methods
        def safe_serialize(obj):
            def default(o):
                return f"{type(o).__qualname__}"

            return json.dumps(obj, default=default)

        return json.loads(safe_serialize(return_dict))



""" Security Related things for the REST-API """
import logging
from functools import wraps

from cryptoadvance.specter.user import User, verify_password as user_verify_password
from flask import current_app as app
from flask import g
from flask_restful import abort

# from flask_httpauth import HTTPBasicAuth

from . import auth

logger = logging.getLogger(__name__)

# auth = HTTPBasicAuth()


@auth.verify_password
def verify_password(username, password):
    """Validate user passwords and store user in the 'g' object"""
    if not username or not password:
        return abort(401)
    the_user = app.specter.user_manager.get_user_by_username(username)
    if not the_user:
        return abort(401)
    g.user = app.specter.user_manager.get_user_by_username(username)
    if user_verify_password(g.user.password, password):
        logger.info(f"Rest-Request for user {username} PASSED password-test")
        return username
    else:
        logger.info(f"Rest-Request for user {username} FAILED password-test")
        return abort(401)

    return g.user is not None and verify_password(g.user.password, password)


def require_admin(func):
    """User needs Admin-rights method decorator"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        """this needs to get implemented properly"""
        # Verify if User is Admin
        if g.user == None:
            return abort(401)
        if g.user.is_admin:
            return func(*args, **kwargs)
        else:
            return abort(401)

    return wrapper



from . import DeviceTypes
from ..device import Device


class SeedSignerDevice(Device):
    device_type = DeviceTypes.SEEDSIGNER
    name = "SeedSigner"
    icon = "seedsigner_icon.svg"

    sd_card_support = False
    qr_code_support = True
    qr_code_frame_rate = 2  # 500 ms per frame
    qr_code_animate = "on"

    def create_psbts(self, base64_psbt, wallet):
        # in QR codes keep only xpubs
        qr_psbt = wallet.fill_psbt(base64_psbt, non_witness=False, xpubs=True)
        # in SD card put as much as possible
        psbts = {"qrcode": qr_psbt}
        return psbts



from collections import OrderedDict
from binascii import hexlify
from embit import base58
from .util.xpub import get_xpub_fingerprint


purposes = OrderedDict(
    {
        "": "General",
        "wpkh": "Single (Segwit)",
        "sh-wpkh": "Single (Nested)",
        "wsh": "Multisig (Segwit)",
        "sh-wsh": "Multisig (Nested)",
        "tr": "Taproot",
    }
)

VALID_PREFIXES = {
    b"\x04\x35\x87\xcf": {  # testnet
        b"\x04\x35\x87\xcf": "",  # unknown, maybe pkh or taproot
        b"\x04\x4a\x52\x62": "sh-wpkh",
        b"\x04\x5f\x1c\xf6": "wpkh",
        b"\x02\x42\x89\xef": "sh-wsh",
        b"\x02\x57\x54\x83": "wsh",
    },
    b"\x04\x88\xb2\x1e": {  # mainnet
        b"\x04\x88\xb2\x1e": "",  # unknown, maybe pkh or taproot
        b"\x04\x9d\x7c\xb2": "sh-wpkh",
        b"\x04\xb2\x47\x46": "wpkh",
        b"\x02\x95\xb4\x3f": "sh-wsh",
        b"\x02\xaa\x7e\xd3": "wsh",
    },
}


class Key:
    def __init__(self, original, fingerprint, derivation, key_type, purpose, xpub):
        if key_type is None:
            key_type = ""
        if fingerprint is None or fingerprint == "":
            fingerprint = get_xpub_fingerprint(original).hex()
        if derivation is None:
            derivation = ""
        if key_type not in purposes:
            key_type = ""
        if not purpose:
            purpose = purposes.get(key_type, "General")
        self.original = original
        self.fingerprint = fingerprint
        self.derivation = derivation
        self.key_type = key_type
        self.purpose = purpose
        self.xpub = xpub

    @classmethod
    def from_json(cls, key_dict):
        original = key_dict.get("original", "")
        fingerprint = key_dict.get("fingerprint", "")
        derivation = key_dict.get("derivation", "")
        key_type = key_dict.get("type", "")
        purpose = key_dict.get("purpose", "")
        xpub = key_dict.get("xpub", "")
        return cls(original, fingerprint, derivation, key_type, purpose, xpub)

    @classmethod
    def parse_xpub(cls, xpub, purpose=""):
        derivation = ""
        arr = xpub.strip().split("]")
        original = arr[-1]
        if len(arr) > 1:
            derivation = arr[0].replace("'", "h").lower()
            xpub = arr[1]

        fingerprint = ""
        # just to be sure fgp/1h/2/3/ is also parsed correctly
        # because we have free-form inputs
        derivation = derivation.rstrip("/")
        if derivation != "":
            if derivation[0] != "[":
                raise Exception("Missing leading [")
            derivation_path = derivation[1:].split("/")
            try:
                fng = bytes.fromhex(
                    derivation_path[0].replace("-", "")
                )  # coldcard has hexstrings like 7c-2c-8e-1b
            except Exception:
                raise Exception("Fingerprint is not hex")
            if len(fng) != 4:
                raise Exception("Incorrect fingerprint length")
            fingerprint = derivation_path[0]
            if len(derivation_path) > 1:
                for path in derivation_path[1:]:
                    if path[-1] == "h":
                        path = path[:-1]
                    try:
                        i = int(path)
                    except:
                        raise Exception("Incorrect index")
                    derivation_path[0] = "m"
                    derivation = "/".join(derivation_path)
            else:
                derivation = ""

        # checking xpub prefix and defining key type
        xpub_bytes = base58.decode_check(xpub)
        prefix = xpub_bytes[:4]
        is_valid = False
        key_type = ""
        for k in VALID_PREFIXES:
            if prefix in VALID_PREFIXES[k].keys():
                key_type = VALID_PREFIXES[k][prefix]
                prefix = k
                is_valid = True
                break
        if not is_valid:
            raise Exception("Invalid xpub prefix: %s", prefix.hex())

        xpub_bytes = prefix + xpub_bytes[4:]
        xpub = base58.encode_check(xpub_bytes)

        # defining key type from derivation
        if derivation != "" and key_type == "":
            derivation_path = derivation.split("/")
            derivation_type = derivation_path[1]
            if derivation_type == "49h":
                key_type = "sh-wpkh"
            elif derivation_type == "84h":
                key_type = "wpkh"
            elif derivation_type == "86h":
                key_type = "tr"
            elif derivation_type == "48h":
                if len(derivation_path) >= 5:
                    if derivation_path[4] == "1h":
                        key_type = "sh-wsh"
                    elif derivation_path[4] == "2h":
                        key_type = "wsh"

        # infer fingerprint and derivation if depth == 0 or depth == 1
        xpub_bytes = base58.decode_check(xpub)
        depth = xpub_bytes[4]
        if depth == 0:
            fingerprint = hexlify(get_xpub_fingerprint(xpub)).decode()
            derivation = "m"
        elif depth == 1:
            fingerprint = hexlify(xpub_bytes[5:9]).decode()
            index = int.from_bytes(xpub_bytes[9:13], "big")
            is_hardened = bool(index & 0x8000_0000)
            derivation = "m/%d%s" % (index & 0x7FFF_FFFF, "h" if is_hardened else "")

        return cls(original, fingerprint, derivation, key_type, purpose, xpub)

    @classmethod
    def parse_xpubs(cls, xpubs):
        xpubs = xpubs
        lines = [l.strip() for l in xpubs.split("\n") if len(l) > 0]
        failed = []
        keys = []
        for line in lines:
            try:
                keys.append(Key.parse_xpub(line))
            except Exception as e:
                failed.append(line + "\n" + str(e))
        return keys, failed

    @property
    def metadata(self):
        metadata = {}
        metadata["chain"] = "Mainnet" if self.xpub.startswith("xpub") else "Testnet"
        metadata["purpose"] = self.purpose
        if self.derivation is not None:
            metadata["combined"] = "[%s%s]%s" % (
                self.fingerprint,
                self.derivation[1:],
                self.xpub,
            )
        else:
            metadata["combined"] = self.xpub
        return metadata

    @property
    def is_testnet(self):
        return not self.xpub.startswith("xpub")

    @property
    def json(self):
        return {
            "original": self.original,
            "fingerprint": self.fingerprint,
            "derivation": self.derivation,
            "type": self.key_type,
            "purpose": self.purpose,
            "xpub": self.xpub,
        }

    def to_string(self, slip132=True):
        if self.derivation and self.fingerprint:
            path_str = f"/{self.derivation[2:]}" if self.derivation != "m" else ""
            return f"[{self.fingerprint}{path_str}]{self.original if slip132 else self.xpub}"
        else:
            return self.original if slip132 else self.xpub

    def __str__(self):
        return self.to_string()

    def __eq__(self, other):
        return self.original == other.original

    def __hash__(self):
        return hash(self.original)



import os, time, requests, platform, tarfile, zipfile, sys, subprocess, shutil, stat, zipfile, logging
import pgpy
from pathlib import Path
from .sha256sum import sha256sum
from .file_download import download_file

logger = logging.getLogger(__name__)


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def setup_tor_thread(specter=None):
    try:
        # There is no Tor Browser binary for Raspberry Pi 4 (armv7l)
        if platform.system() == "Linux" and "armv" in platform.machine():
            raise Exception(
                "Linux ARM devices (e.g. Raspberry Pi) must manually install Tor"
            )

        specter.update_setup_status("torbrowser", "STARTING_SETUP")
        TOR_OS_SUFFIX = {
            "Windows": "tor-win64-0.4.5.7.zip",
            "Linux": "tor-browser-linux64-10.0.15_en-US.tar.xz",
            "Darwin": "TorBrowser-10.0.15-osx64_en-US.dmg",
        }

        packed_name = (
            os.path.join(sys._MEIPASS, f"torbrowser/{TOR_OS_SUFFIX[platform.system()]}")
            if getattr(sys, "frozen", False)
            else Path(__file__).parent
            / f"../../../../pyinstaller/torbrowser/{TOR_OS_SUFFIX[platform.system()]}"
        )

        os.makedirs(os.path.join(specter.data_folder, "tor-binaries"), exist_ok=True)
        if TOR_OS_SUFFIX[platform.system()].endswith(".dmg"):
            result = subprocess.run(
                [f'hdiutil attach "{packed_name}"'], shell=True, capture_output=True
            )
            if result.stderr == b"hdiutil: attach failed - no mountable file systems\n":
                os.remove(packed_name)
                return
            torbrowser_source_path = os.path.join(
                "/Volumes", "Tor Browser", "Tor Browser.app", "Contents", "MacOS", "Tor"
            )
            copytree(
                src=torbrowser_source_path,
                dst=os.path.join(specter.data_folder, "tor-binaries"),
            )
            disk_image_path = "/Volumes/Tor\ Browser"
            subprocess.run([f"hdiutil detach {disk_image_path}"], shell=True)
        elif TOR_OS_SUFFIX[platform.system()].endswith(".zip"):
            with zipfile.ZipFile(packed_name) as zip_file:
                for file in zip_file.filelist:
                    if file.filename.endswith("dll") or file.filename.endswith("exe"):
                        destination_exe = os.path.join(
                            os.path.join(specter.data_folder, "tor-binaries"),
                            file.filename.split("/")[-1],
                        )
                        with zip_file.open(file.filename) as zf, open(
                            destination_exe, "wb"
                        ) as f:
                            shutil.copyfileobj(zf, f)
        elif TOR_OS_SUFFIX[platform.system()].endswith(".tar.xz"):
            with tarfile.open(packed_name) as tar:
                tor_files = [
                    "libcrypto.so.1.1",
                    "libevent-2.1.so.7",
                    "libssl.so.1.1",
                    "tor",
                ]
                for tor_file in tor_files:
                    file_name = "tor-browser_en-US/Browser/TorBrowser/Tor/" + tor_file
                    destination_file = os.path.join(
                        os.path.join(specter.data_folder, "tor-binaries"), tor_file
                    )
                    extracted_file = tar.extractfile(file_name)
                    with open(destination_file, "wb") as f:
                        shutil.copyfileobj(extracted_file, f)
                        if tor_file == "tor":
                            st = os.stat(destination_file)
                            os.chmod(destination_file, st.st_mode | stat.S_IEXEC)

        if "torrc_password" not in specter.config:
            specter.generate_torrc_password()
        with open(os.path.join(specter.data_folder, "torrc"), "w") as file:
            file.write("ControlPort 9051")
            file.write(
                f"\nHashedControlPassword {specter.tor_daemon.get_hashed_password(specter.config['torrc_password'])}"
            )

        specter.tor_daemon.start_tor_daemon()
        specter.update_tor_controller()
        specter.reset_setup("torbrowser")
    except Exception as e:
        logger.error(f"Failed to install Tor. Error: {e}")
        specter.update_setup_error("torbrowser", str(e))



import logging
import os
import shutil
import signal
from pathlib import Path

import click
import psutil

logger = logging.getLogger(__name__)


class Echo:
    def __init__(self, quiet):
        self.quiet = quiet

    def echo(self, mystring, prefix=True, **kwargs):
        if self.quiet:
            pass
        else:
            if prefix:
                click.echo(f"    --> ", nl=False)
            click.echo(f"{mystring}", **kwargs)


def kill_node_process(node_impl, echo):
    did_something = False
    for proc in psutil.process_iter():
        try:
            # Get process name & pid from process object.
            if proc.name().endswith(f"{node_impl}d"):
                echo(f"Killing {node_impl}d-process with id {proc.pid} ...")
                did_something = True
                os.kill(proc.pid, signal.SIGTERM)
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            echo(f"Pid {proc.pid} not owned by us. Might be a docker-process? {proc}")
    return did_something


def compute_data_dir_and_set_config_obj(node_impl, data_dir, config_obj):
    if node_impl == "bitcoin":
        if data_dir:
            config_obj["BTCD_REGTEST_DATA_DIR"] = data_dir
        return config_obj["BTCD_REGTEST_DATA_DIR"]
    elif node_impl == "elements":
        if data_dir:
            config_obj["ELMD_REGTEST_DATA_DIR"] = data_dir
        return config_obj["ELMD_REGTEST_DATA_DIR"]


def purge_node_data_dir(node_impl, config_obj, echo):
    did_something = False
    if node_impl == "elements":
        data_dir = config_obj["ELMD_REGTEST_DATA_DIR"]
    elif node_impl == "bitcoin":
        data_dir = config_obj["BTCD_REGTEST_DATA_DIR"]
    else:
        raise Exception(f"unknown node_impl {node_impl}")

    if Path(data_dir).exists():
        echo(f"Purging Datadirectory {data_dir} ...")
        did_something = True
        shutil.rmtree(data_dir)
    return did_something



from .node_controller import NodePlainController
import tempfile
import logging

logger = logging.getLogger(__name__)


class ElementsPlainController(NodePlainController):
    """A Controller specifically for the Bitcoind-process"""

    def __init__(
        self,
        elementsd_path="elementsd",
        rpcport=18555,
        network="regtest",
        rpcuser="liquid",
        rpcpassword="secret",
    ):
        # Just call super and add the node_impl
        super().__init__(
            "elements",
            node_path=elementsd_path,
            rpcport=rpcport,
            network=network,
            rpcuser=rpcuser,
            rpcpassword=rpcpassword,
        )

    def start_elementsd(
        self,
        cleanup_at_exit=False,
        cleanup_hard=False,
        datadir=None,
        log_stdout=False,
        extra_args=[],
        timeout=60,
    ):
        """starts elementsd with a specific rpcport=18543 by default.
        That's not the standard in order to make pytest running while
        developing locally against a different regtest-instance
        """
        # convenience method
        return self.start_node(
            cleanup_at_exit,
            cleanup_hard,
            datadir,
            log_stdout,
            extra_args,
            timeout,
        )

    def stop_elementsd(self):
        self.stop_node()

    @classmethod
    def construct_node_cmd(
        cls,
        rpcconn,
        run_docker=True,
        datadir=None,
        log_stdout=False,
        node_path="elementsd",
        network="regtest",  # Doesn't make sense here. For now, only "elreg" is supported
        extra_args=[],
    ):
        """returns a command to run your elementsd"""
        btcd_cmd = '"{}" '.format(node_path)
        btcd_cmd += " -chain=elreg "
        btcd_cmd += " -fallbackfee=0.0000001 "
        btcd_cmd += " -validatepegin=0 "
        btcd_cmd += " -txindex=1 "
        btcd_cmd += " -initialfreecoins=2100000000000000 "
        btcd_cmd += " -port={} -rpcport={} -rpcbind=0.0.0.0".format(
            rpcconn.rpcport + 1, rpcconn.rpcport
        )
        btcd_cmd += " -rpcuser={} -rpcpassword={} ".format(
            rpcconn.rpcuser, rpcconn.rpcpassword
        )
        btcd_cmd += " -rpcallowip=0.0.0.0/0 -rpcallowip=172.17.0.0/16 "
        if not run_docker:
            if not log_stdout:
                btcd_cmd += " -noprinttoconsole"
            if datadir == None:
                datadir = tempfile.mkdtemp(prefix="bitcoind_datadir")
            btcd_cmd += ' -datadir="{}" '.format(datadir)
        print(f"extra_args={extra_args})")
        if extra_args:
            btcd_cmd += " {}".format(" ".join(extra_args))
        logger.debug("constructed elementsd-command: %s", btcd_cmd)
        return btcd_cmd

    def version(self):
        """Returns the version of elementsd, e.g. "v0.18.1" """
        version = self.get_rpc().getnetworkinfo()["subversion"]
        version = version.replace("/", "").replace("Elements Core:", "v")
        return version



""" A config module contains static configuration """
import configparser
import datetime
import os
import random
import secrets
from pathlib import Path

from dotenv import load_dotenv

# BASEDIR = os.path.abspath(os.path.dirname(__file__))

# Loading env-vars from .flaskenv (4 levels above this file)
env_path = Path("../../../..") / ".flaskenv"
load_dotenv(env_path)


def _get_bool_env_var(varname, default=None):

    value = os.environ.get(varname, default)

    if value is None:
        return False
    elif isinstance(value, str) and value.lower() == "false":
        return False
    elif bool(value) is False:
        return False
    else:
        return bool(value)


DEFAULT_CONFIG = "cryptoadvance.specter.config.DevelopmentConfig"


class BaseConfig(object):
    PORT = os.getenv("PORT", 25441)
    CONNECT_TOR = _get_bool_env_var(os.getenv("CONNECT_TOR", "False"))
    SPECTER_DATA_FOLDER = os.path.expanduser(
        os.getenv("SPECTER_DATA_FOLDER", "~/.specter")
    )
    SPECTER_API_ACTIVE = _get_bool_env_var("SPECTER_API_ACTIVE", "False")
    # Logging
    # SPECTER_LOGFILE will get created dynamically in server.py
    # using:
    SPECTER_LOGFORMAT = os.getenv(
        "SPECTER_LOGFORMAT", "[%(asctime)s] %(levelname)s in %(module)s: %(message)s"
    )

    # CERT and KEY is for running self-signed-ssl-certs. Check cli_server for details
    CERT = os.getenv("CERT", None)
    KEY = os.getenv("KEY", None)
    # This will get passed to initialize the specter-object
    DEFAULT_SPECTER_CONFIG = {}

    # only used by cli_bitcoind.py, we want to have that static for the same reason
    BTCD_REGTEST_DATA_DIR = os.getenv(
        "BTCD_REGTEST_DATA_DIR", "/tmp/specter_btc_regtest_plain_datadir"
    )

    # only used by cli_node.py, we want to have that static for the same reason
    ELMD_REGTEST_DATA_DIR = os.getenv(
        "ELMD_REGTEST_DATA_DIR", "/tmp/specter_elm_regtest_plain_datadir"
    )

    # the default timeout for Bitcoin/Liquid RPC-calls
    BITCOIN_RPC_TIMEOUT = int(os.getenv("BITCOIN_RPC_TIMEOUT", "10"))
    LIQUID_RPC_TIMEOUT = int(os.getenv("LIQUID_RPC_TIMEOUT", "10"))

    # The self-signed ssl-certificate which is lazily created is configurable to a certain extent
    SPECTER_SSL_CERT_SUBJECT_C = os.getenv("SPECTER_SSL_CERT_SUBJECT_C", "DE")
    SPECTER_SSL_CERT_SUBJECT_ST = os.getenv("SPECTER_SSL_CERT_SUBJECT_ST", "BDW")
    SPECTER_SSL_CERT_SUBJECT_L = os.getenv("SPECTER_SSL_CERT_SUBJECT_L", "Freiburg")
    SPECTER_SSL_CERT_SUBJECT_O = os.getenv(
        "SPECTER_SSL_CERT_SUBJECT_O", "Specter Citadel Cert"
    )
    SPECTER_SSL_CERT_SUBJECT_OU = os.getenv(
        "SPECTER_SSL_CERT_SUBJECT_OU", "Specter Citadel Cert"
    )
    SPECTER_SSL_CERT_SUBJECT_CN = os.getenv(
        "SPECTER_SSL_CERT_SUBJECT_CN", "Specter Citadel Cert"
    )
    # For self-signed certs, serial-number collision is a risk, so let's do a random one by default
    SPECTER_SSL_CERT_SERIAL_NUMBER = int(
        os.getenv("SPECTER_SSL_CERT_SERIAL_NUMBER", random.randrange(1, 100000))
    )
    INTERNAL_BITCOIND_VERSION = os.getenv("INTERNAL_BITCOIND_VERSION", "0.21.1")

    # Block explorers URLs
    EXPLORERS_LIST = {
        "MEMPOOL_SPACE": {
            "name": "Mempool.space",
            "url": "https://mempool.space/",
        },
        "MEMPOOL_SPACE_ONION": {
            "name": "Mempool.space Tor hidden service",
            "url": "http://mempoolhqx4isw62xs7abwphsq7ldayuidyx2v2oethdhhj6mlo2r6ad.onion/",
        },
        "BLOCKSTREAM_INFO": {
            "name": "Blockstream.info",
            "url": "https://blockstream.info/",
        },
        "BLOCKSTREAM_INFO_ONION": {
            "name": "Blockstream.info Tor hidden service",
            "url": "http://explorerzydxu5ecjrkwceayqybizmpjjznk5izmitf2modhcusuqlid.onion/",
        },
        "CUSTOM": {
            "name": "Custom",
            "url": "",
        },
    }

    # Babel integration. English listed first; other alphabetical by language code
    LANGUAGES = {
        "en": "English",
        "bg": "Български",
        "de": "Deutsch",
        "el": "Ελληνικά",
        "es": "Español",
        "fr": "Français",
        "he": "עברית",
        "hi": "हिंदी",
        "nl": "Nederlands",
        "pl": "Polski",
        "pt": "Português",
        "ru": "Русский",
        "sv": "Svenska",
        "zh_Hans_CN": "简体中文",
        "zh_Hant_TW": "繁體中文",
    }

    # Babel integration. List of languages written from right to left for RTL support in the UI
    RTL_LANGUAGES = ["he"]


class DevelopmentConfig(BaseConfig):
    # https://stackoverflow.com/questions/22463939/demystify-flask-app-secret-key
    SECRET_KEY = "development key"
    SPECTER_DATA_FOLDER = os.path.expanduser(
        os.getenv("SPECTER_DATA_FOLDER", "~/.specter_dev")
    )
    # API active by default in dev-mode
    SPECTER_API_ACTIVE = _get_bool_env_var("SPECTER_API_ACTIVE", "True")


class TestConfig(BaseConfig):
    SECRET_KEY = "test key"
    # This should never be used as the data-folder is injected at runtime
    # But let's be sure before something horrible happens:
    SPECTER_DATA_FOLDER = os.path.expanduser(
        os.getenv("SPECTER_DATA_FOLDER", "~/.specter_testing")
    )
    # API active by default in test-mode
    SPECTER_API_ACTIVE = _get_bool_env_var("SPECTER_API_ACTIVE", "True")

    # See #1316 since Bitcoin v0.21.1 (not only) the importmulti-call takes longer than 10 seconds on cirrus
    BITCOIN_RPC_TIMEOUT = 20
    LIQUID_RPC_TIMEOUT = 30


class CypressTestConfig(TestConfig):
    SPECTER_DATA_FOLDER = os.path.expanduser(
        os.getenv("SPECTER_DATA_FOLDER", "~/.specter_cypress")
    )
    PORT = os.getenv("PORT", 25444)

    # need to be static in order to (un-)tar bitcoind-dirs reliable
    DEFAULT_SPECTER_CONFIG = {"uid": "123456"}

    BTCD_REGTEST_DATA_DIR = os.getenv(
        "BTCD_REGTEST_DATA_DIR", "/tmp/specter_cypress_btc_regtest_plain_datadir"
    )

    BTCD_REGTEST_DATA_DIR = os.getenv(
        "BTCD_REGTEST_DATA_DIR", "/tmp/specter_cypress_elm_regtest_plain_datadir"
    )
    BITCOIN_RPC_TIMEOUT = 30
    LIQUID_RPC_TIMEOUT = 40


class ProductionConfig(BaseConfig):
    SECRET_KEY = secrets.token_urlsafe(16)



import copy, random, json, os, threading, shutil, logging

from flask import (
    Flask,
    Blueprint,
    render_template,
    request,
    redirect,
    url_for,
    jsonify,
    flash,
)
from flask_babel import lazy_gettext as _
from flask_login import login_required, current_user
from flask import current_app as app
from mnemonic import Mnemonic
from ..helpers import alias, is_testnet, generate_mnemonic
from ..key import Key
from ..managers.device_manager import get_device_class
from ..devices.bitcoin_core import BitcoinCore
from ..managers.wallet_manager import purposes
from ..specter_error import handle_exception
from ..util.bitcoind_setup_tasks import (
    setup_bitcoind_thread,
    setup_bitcoind_directory_thread,
)
from ..util.tor_setup_tasks import (
    setup_tor_thread,
)

logger = logging.getLogger(__name__)

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
setup_endpoint = Blueprint("setup_endpoint", __name__)

######################### Setup pages #######################################
@setup_endpoint.route("/start/", methods=["GET"])
@login_required
def start():
    app.specter.setup_status["stage"] = "start"
    return render_template("setup/start.jinja", specter=app.specter, rand=rand)


@setup_endpoint.route("/tor/", methods=["GET"])
@login_required
def tor():
    """wizard: Setup Tor daemon (Skip / Setup Tor)"""
    app.specter.setup_status["stage"] = "tor"
    return render_template(
        "setup/tor.jinja",
        nextURL="setup_endpoint.node_type",
        specter=app.specter,
        rand=rand,
    )


@setup_endpoint.route("/node_type/", methods=["GET"])
@login_required
def node_type():
    """wizard: Would you like to setup a new Bitcoin node or connect to an existing one?
    (Connect existing node / Setup a new node )
    """
    app.specter.setup_status["stage"] = "node_type"
    return render_template(
        "setup/node_type.jinja",
        bitcoind_installed=os.path.isfile(app.specter.bitcoind_path),
        specter=app.specter,
        rand=rand,
    )


@setup_endpoint.route("/bitcoind/", methods=["GET"])
@login_required
def bitcoind():
    """wizard: Setup Bitcoin Core (Start the Setup!)"""
    app.specter.setup_status["stage"] = "bitcoind"
    return render_template("setup/bitcoind.jinja", specter=app.specter, rand=rand)


@setup_endpoint.route(
    "/bitcoind_datadir/", defaults={"network": "main"}, methods=["GET"]
)
@setup_endpoint.route("/bitcoind_datadir/<network>", methods=["GET"])
@login_required
def bitcoind_datadir(network):
    """wizard: Configure your node (Quicksync? , Start Syncing)"""
    app.specter.setup_status["stage"] = "bitcoind_datadir"
    return render_template(
        "setup/bitcoind_datadir.jinja", network=network, specter=app.specter, rand=rand
    )


@setup_endpoint.route("/end/", methods=["GET"])
@login_required
def end():
    """wizard: Setup competed Successfully (Done)"""
    app.specter.setup_status["stage"] = "start"
    return render_template("setup/end.jinja", specter=app.specter, rand=rand)


@setup_endpoint.route("/tor_setup/", methods=["GET"])
@login_required
def tor_from_settings():
    """wizard: Setup Tor daemon (Setup Tor)"""
    return render_template(
        "setup/tor.jinja",
        nextURL="settings_endpoint.tor",
        specter=app.specter,
        rand=rand,
    )


######################### Background tasks #######################################
@setup_endpoint.route("/setup_tor", methods=["POST"])
@login_required
def setup_tor():
    if (
        not os.path.isfile(app.specter.torbrowser_path)
        and app.specter.setup_status["torbrowser"]["stage_progress"] == -1
    ):
        t = threading.Thread(target=setup_tor_thread, args=(app.specter,))
        t.start()
    elif os.path.isfile(app.specter.torbrowser_path):
        return {"error": _("Tor is already installed")}
    elif app.specter.setup_status["torbrowser"]["stage_progress"] != -1:
        return {"error": _("Tor installation is still in progress")}
    return {"success": _("Starting Tor setup!")}


@setup_endpoint.route("/setup_bitcoind", methods=["POST"])
@login_required
def setup_bitcoind():
    if (
        not os.path.isfile(app.specter.bitcoind_path)
        and app.specter.setup_status["bitcoind"]["stage_progress"] == -1
    ):
        app.specter.update_setup_status("bitcoind", "STARTING_SETUP")
        t = threading.Thread(
            target=setup_bitcoind_thread,
            args=(app.specter, app.config["INTERNAL_BITCOIND_VERSION"]),
        )
        t.start()
    elif os.path.isfile(app.specter.bitcoind_path):
        return {"error": _("Bitcoin Core is already installed")}
    elif app.specter.setup_status["bitcoind"]["stage_progress"] != -1:
        return {"error": _("Bitcoin Core installation is still in progress")}
    return {"success": _("Starting Bitcoin Core setup!")}


@setup_endpoint.route("/setup_bitcoind_datadir", methods=["POST"])
@login_required
def setup_bitcoind_datadir():
    network = request.form.get("network", "main")
    node_name = "Specter Bitcoin" if network == "main" else f"Specter {network.title()}"
    i = 1
    while node_name in app.specter.node_manager.nodes:
        i += 1
        node_name = (
            f"Specter Bitcoin {i}"
            if network == "main"
            else f"Specter {network.title()} {i}"
        )
    node_default_datadir = os.path.join(
        app.specter.node_manager.data_folder, f"{alias(node_name)}/.bitcoin-{network}"
    )
    user_selected_datadir = request.form.get(
        "bitcoin_core_datadir", node_default_datadir
    )
    if os.path.exists(user_selected_datadir):
        if request.form["override_data_folder"] != "true":
            logger.warning(
                f"Bitcoin Core data directory at {user_selected_datadir} already exists and no orride permission was explicitly given."
            )
            return {"error": "data folder already exists"}
        logger.info(f"Deleting Bitcoin Core data directory at: {user_selected_datadir}")
        shutil.rmtree(user_selected_datadir, ignore_errors=True)
    if (
        os.path.isfile(app.specter.bitcoind_path)
        and app.specter.setup_status["bitcoind"]["stage_progress"] == -1
    ):
        node = app.specter.node_manager.add_internal_node(
            node_name,
            network=network,
            datadir=user_selected_datadir,
        )
        app.specter.update_setup_status("bitcoind", "STARTING_SETUP")
        quicksync = request.form["quicksync"] == "true"
        pruned = request.form["nodetype"] == "pruned"
        t = threading.Thread(
            target=setup_bitcoind_directory_thread,
            args=(app.specter, quicksync, pruned, node.alias),
        )
        t.start()
    elif not os.path.isfile(app.specter.bitcoind_path):
        return {"error": _("Bitcoin Core in not installed but required for this step")}
    elif app.specter.setup_status["bitcoind"]["stage_progress"] != -1:
        return {"error": _("Bitcoin Core installation is still in progress")}
    return {"success": _("Starting Bitcoin Core setup!")}


######################### Setup status (open endpoint) #######################################
@setup_endpoint.route("/get_software_setup_status/<software>")
@login_required
@app.csrf.exempt
def get_software_setup_status(software):
    return app.specter.get_setup_status(software)
